-- Functions that help coding cards with "can't untap more than..." effects, such as Static Orb

TFM_UL_CARD_TYPE_PERMANENT = -1111
TFM_UL_UNLIMITED = 9999
TFM_UL_PERMANENT_TYPES = {
	TFM_UL_CARD_TYPE_PERMANENT,
	CARD_TYPE_ARTIFACT,
	CARD_TYPE_CREATURE,
	CARD_TYPE_ENCHANTMENT,
	CARD_TYPE_LAND,
	CARD_TYPE_PLANESWALKER,
	CARD_TYPE_TRIBAL
}

TFM_SetUntapLimit = function(amount, type, player)
-- To be used in a trigger that fires before the player's untap step, sets that player's untap limit for a card type.
-- If no type is specified (or TFM_UL_CARD_TYPE_PERMANENT is used), the limit applies to all permanents.

	-- Send the untap limit to the manager
	amount = amount or 1
	type = type or TFM_UL_CARD_TYPE_PERMANENT
	player = player or TriggerPlayer()
	local playerDC = player:PlayerDataChest()
	local DC1 = TFM_GetOrMakeChest(playerDC, "TFM_Untap_Limits_1")
	local subDC = DC1:Make_Chest( DC1:Count() )
	subDC:Set_Int(0, type)
	subDC:Set_Int(1, amount)

	MTG():ClearFilterMark() -- Prepare to mark untappable cards (the BECAME_UNTAPPED trigger should mark them)

	MTG():DuelDataChest():Set_Int("TFM_Untap_Limits_On", 1) -- Turn on the override of the untapping
end

TFM_ManageUntapLimit1 = function(player)
-- To be used in an untap step beginning trigger, reads the untap limits for the given player and reorganizes them.
	player = player or TriggerPlayer()
	local playerDC = player:PlayerDataChest()
	local DC1 = playerDC:Get_Chest("TFM_Untap_Limits_1") -- This chest has the untap limits sent by the cards
	if DC1 == nil then
		return -- This player has no limits: do nothing
	end
	playerDC:Free_Compartment("TFM_Untap_Limits_2")
	local DC2 = playerDC:Make_Chest("TFM_Untap_Limits_2") -- This chest will have the reorganized untap limits
	for i, type in ipairs(TFM_UL_PERMANENT_TYPES) do
		DC2:Set_Int(type, TFM_UL_UNLIMITED) -- Any type with this large number will be recognized as having no limit
	end

	-- Get each limit sent by the cards
	local count = DC1:Count()
	for i=0,count-1 do
		local limitDC = DC1:Get_Chest(i)
		local type = limitDC:Get_Int(0)
		local amount = limitDC:Get_Int(1)
		local currentAmount = DC2:Get_Int(type)
		if amount < currentAmount then
			DC2:Set_Int(type, amount) -- There's a lower limit for this type: remember it
		end
	end
	playerDC:Free_Compartment("TFM_Untap_Limits_1") -- Delete the limits sent by the cards, we don't need them anymore
end

TFM_ManageUntapLimit2 = function(player)
-- To be used after TFM_ManageUntapLimit1 and in a repeating action, queries the given player on what to untap.
-- Remember to write "return" before the function: the repeating action needs to know whether it should repeat or not!
	player = player or TriggerPlayer()
	local playerDC = player:PlayerDataChest()
	local DC2 = playerDC:Get_Chest("TFM_Untap_Limits_2")
	if DC2 == nil then
		return false -- This player has no limits: exit
	end
	local n = MTG():GetActionRepCount()
	local filter = ClearFilter()
	if n>0 then
		local lastTarget = EffectDC():Get_Targets(n-1) and EffectDC():Get_Targets(n-1):Get_CardPtr(0)
		if lastTarget ~= nil then
			-- Subtract the last target's types from the limit
			local lastTargetTypes = lastTarget:GetCardType()
			for i, type in ipairs(TFM_UL_PERMANENT_TYPES) do
				if DC2:Get_Int(type) < TFM_UL_UNLIMITED and
				( type == TFM_UL_CARD_TYPE_PERMANENT or lastTargetTypes:Test(type) ) then
					DC2:Int_Sub(type, 1)
				end
			end
		end
		for i=0,n-1 do
			local target = EffectDC():Get_Targets(i) and EffectDC():Get_Targets(i):Get_CardPtr(0)
			if target ~= nil then
				filter:Add(FE_CARD_INSTANCE, OP_NOT, target) -- Disallow selecting a previous target again
			end
		end
	end
	if DC2:Get_Int(TFM_UL_CARD_TYPE_PERMANENT) == 0 then
		return false -- No more permanents to untap: exit
	end
	filter:Add(FE_CONTROLLER, OP_IS, player)
	filter:Add(FE_IS_TAPPED, true)
	local subFilter = nil
	if DC2:Get_Int(TFM_UL_CARD_TYPE_PERMANENT) == TFM_UL_UNLIMITED then
		subFilter = filter:AddSubFilter_Or() -- No limit on all permanents, but only on specific types
	end
	for i, type in ipairs(TFM_UL_PERMANENT_TYPES) do
		if type ~= TFM_UL_CARD_TYPE_PERMANENT then
			if DC2:Get_Int(type) == 0 then
				filter:Add(FE_TYPE, OP_NOT, type) -- Disallow types whose limit went to 0
			elseif subFilter ~= nil and DC2:Get_Int(type) < TFM_UL_UNLIMITED then
				subFilter:Add(FE_TYPE, OP_IS, type) -- Restrict to specific types
			end
		end
	end
	filter:SetMarkedObjectsOnly() -- Only untappable cards can be selected
	if filter:CountStopAt(1) == 0 then
		return false -- No feasible targets: exit
	end
	player:ChooseItem( "CARD_QUERY_CHOOSE_PERMANENT_TO_UNTAP", EffectDC():Make_Targets(n) )
	return true
end

TFM_ManageUntapLimit3 = function()
-- To be used after TFM_ManageUntapLimit2, untaps what has been selected.
	MTG():DuelDataChest():Set_Int("TFM_Untap_Limits_On", 0) -- Turn off the override of the untapping
	local i = 0
	local target = EffectDC():Get_Targets(i) and EffectDC():Get_Targets(i):Get_CardPtr(0)
	while target ~= nil do
		target:Untap()
		i = i+1
		target = EffectDC():Get_Targets(i) and EffectDC():Get_Targets(i):Get_CardPtr(0)
	end
end